#include <errno.h>
#include "tcpip.h"

proc_matcher_find matcher_find;
proc_alert alert;

void iprule_cleanup(struct ip_rule *r)
{
	if ( !r ) return;

	free(r);
}

int iprule_addcrit(struct criteria *c, struct ip_rule *rule, struct criteria *mods)
{
	struct relement *rel;

	/* Alocate a new criteria */
	if ( !(rel=malloc(sizeof(*rel))) ) {
		return 0;
	}

	/* Find the right matcher to use */
	if ( !(rel->match=matcher_find(c->crit)) ) {
		free(rel);
		return 0;
	}

	/* Validate that shit */
	if ( !(rel->m=rel->match->validate(c->args, &rel->priv, mods)) ) {
		free(rel);
		return 0;
	}

	rel->negate=c->negate;
	rel->next=NULL;

	/* Link it on to the rule */
	if ( !rule->criteria ) {
		rule->criteria=rel;
		rule->criteria_head=rel;
	}else{
		rule->criteria_head->next=rel;
		rule->criteria_head=rel;
	}

	return 1;
}

/* Link a header node on to a list */
struct tcp_head *tcp_udp_rlink(struct tcp_head *n,
		struct tcp_head **list,
		unsigned int *count)
{
	struct tcp_head *tmp;

	for(tmp=*list; tmp; tmp=tmp->next) {
		if ( (tmp->saddr==n->saddr) &&
			(tmp->daddr==n->daddr) &&
			(tmp->smask==n->smask) &&
			(tmp->dmask==n->dmask) &&
			(tmp->sport_min==n->sport_min) &&
			(tmp->sport_max==n->sport_max) &&
			(tmp->dport_min==n->dport_min) &&
			(tmp->dport_max==n->dport_max) &&
			(tmp->flags==n->flags) ) {
			free(n);
			return tmp;
		}
	}

	*count+=1;
	n->next=*list;
	*list=n;
	return n;
}

int tcp_udp_commit(struct tcp_head **tr,
	unsigned int *hcount,
	unsigned int *rcount,
	struct rule *r)
{
	struct criteria *c;
	struct ip_rule *rule;
	struct tcp_head *hd;
	int err=0;

	if ( !(hd=calloc(1, sizeof(*hd))) ) {
		return 0;
	}
	
	if ( !(rule=calloc(1, sizeof(*rule))) ) {
		free(hd);
		return 0;
	}

	hd->sport_max=0xffff;
	hd->dport_max=0xffff;

	for(c=r->criteria; c; c=c->next) {
		if ( !strcmp("src", c->crit) ) {
			if ( c->negate ) hd->flags|=NEGATE_SADDR;
			if ( !parse_ipaddr(c->args, &hd->saddr, &hd->smask) ) {
				err=1;
				break;
			}
			continue;
		}

		if ( !strcmp("dst", c->crit) ) {
			if ( c->negate ) hd->flags|=NEGATE_DADDR;
			if ( !parse_ipaddr(c->args, &hd->daddr, &hd->dmask) ) {
				err=1;
				break;
			}
			continue;
		}

		if ( !strcmp("dport", c->crit) ) {
			if ( c->negate ) hd->flags|=NEGATE_DPORT;
			if ( !parse_ports(c->args, &hd->dport_min, &hd->dport_max) ) {
				err=1;
				break;
			}
			continue;
		}

		if ( !strcmp("sport", c->crit) ) {
			if ( c->negate ) hd->flags|=NEGATE_SPORT;
			if ( !parse_ports(c->args, &hd->sport_min, &hd->sport_max) ) {
				err=1;
				break;
			}
			continue;
		}

		if ( !strcmp("ip_proto", c->crit) ) {
			err=1;
			break;
		}

		if ( !iprule_addcrit(c, rule, r->modifier) ) {
			err=1;
			break;
		}
	}

	if ( err ) {
		iprule_cleanup(rule);
		free(hd);
		return 0;
	}

	/* Link in header node, can't fail */
	hd=tcp_udp_rlink(hd, tr, hcount);

	/* Finish off rule stuff */
	memcpy(&rule->a, &r->alert, sizeof(rule->a));

	/* Link rule on to rule header */
	rule->next=hd->rules;
	hd->rules=rule;
	*rcount+=1;
	
	return 1;
}

/* May seem a bit over the top but I want to be very strict
 * about what is valid and what isn't */
int parse_ports(char *str, u_int16_t *min, u_int16_t *max)
{
	unsigned long int num;
	char *lo=NULL, *hi=NULL;
	char *tmp;
	int state=0;

	/* Parse the string out */
	for(tmp=str; *tmp; tmp++) {
		if ( *tmp!=':' && (*tmp<'0'||*tmp>'9') )
			return 0;
		
		if ( state==0 ) {
			if ( *tmp==':' ) {
				state=1;
				*tmp=0;
			}else if ( !lo && (*tmp>='0' && *tmp<='9') ) {
				lo=tmp;
			}
		}else if ( state==1 ) {
			if ( *tmp>='0' && *tmp<='9' ) {
				hi=tmp;
				break;
			}
		}
	}
	
	/* Low part is easy */
	if ( lo ) {
		if ( (num=strtoul(lo, NULL, 10))==ULONG_MAX &&
			errno==ERANGE ) return 0;
		if ( num & ~0xffff ) return 0;

		*min=num&0xffff;
	} else *min=0;
	
	if ( hi ) {
		if ( (num=strtoul(hi, NULL, 10))==ULONG_MAX &&
			errno==ERANGE ) return 0;
		if ( num & ~0xffff ) return 0;
		
		*max=num&0xffff;
	}else{
		/* If state is 1 that means there was a colon
		 * eg: 1024: otherwise it was just a single
		 * number, eg: '23' in which case max
		 * is just equal to min */
		if ( state==1 ) 
			*max=0xffff;
		else
			*max=*min;
	}

	return 1;
}

int parse_ipproto(char *str, u_int8_t *prot)
{
	unsigned long int num;

	if ( (num=strtoul(str, NULL, 10))==ULONG_MAX &&
		errno==ERANGE ) {
		return 0;
	}

	/* check we're not out of range */
	if ( num & ~(0xff) ) return 0;
	
	*prot=(u_int8_t)num & 0xff;
	return 1;
}

int parse_ipaddr(char *str, u_int32_t *addr, u_int32_t *mask)
{
	char *smask;
	struct in_addr tmp;

	if ( (smask=strchr(str, '/')) ) {
		*smask='\0'; smask++;

		if ( strchr(smask, '.') ) { 
			/* Dotted style mask notation (255.255.255.0) */
			if ( !inet_aton(smask, &tmp) ) return 0;
			*mask=tmp.s_addr;
		}else{ /* CIDR notation (/24) */
			int bits=atoi(smask);
			if ( bits > 32 || bits < 0 ) return 0;
			*mask=~(-1 << bits);
		}
	}else{
		*mask=-1;
	}

	if ( !inet_aton(str, &tmp) ) return 0;
	*addr=tmp.s_addr;

	return 1;
}

int _PLUGIN_INIT (struct plugin_in *in, struct plugin_out *out)
{
	proc_decode_add decode_add;
	proc_realm_add realm_add;
	proc_realm_add realm_add2;
	unsigned int ok=0;
	
	plugin_check(in, out);
	
	PLUGIN_ID("decode.tcpip", "The Internet Protocol");
	PLUGIN_VERSION(1, 0);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@ecsc.co.uk");
	PLUGIN_LICENSE("GPL");
	
	if ( !(decode_add=in->import("decode.add")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(realm_add=in->import("realm.add")) ) {
		return PLUGIN_ERR_OBJECT;
	}
	
	if ( !(realm_add2=in->import("realm.add2")) ) {
		return PLUGIN_ERR_OBJECT;
	}
	
	if ( !(matcher_find=in->import("matcher.find")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(alert=in->import("alert")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	ok+=decode_add(&ipv4_p, ipv4_r);
	ok+=decode_add(&icmp_p, icmp_r);
	ok+=decode_add(&tcp_p, tcp_r);
	ok+=decode_add(&udp_p, udp_r);

	if ( (ip_realm=realm_add("ip", ip_match, ip_commit)) ) ok++;

	/* These ones piggyback IP so don't need an entry in the
	 * realms bitmask, hence realm_add2 */
	if ( (tcp_realm=realm_add2("tcp", tcp_match, tcp_commit)) ) ok++;
	if ( (udp_realm=realm_add2("udp", udp_match, udp_commit)) ) ok++;
	if ( (icmp_realm=realm_add2("icmp", icmp_match, icmp_commit)) ) ok++;
	
	return (ok) ? PLUGIN_ERR_OK : PLUGIN_ERR_FAIL;
}

int _PLUGIN_UNLOAD (int code) {
	/* TODO: Cleanup realms */
	return PLUGIN_ERR_OK;
}
